﻿

'
'/**
'--------------------------------------------------
' Version .....: 02/01/2011
'               [dd/mm/aaaa] 
' Original en .: Util
'--------------------------------------------------
'
'
' -------------------------------------------------
' Clase ....: SingletonBaseGenerica
' Lenguaje .: (.NET Framework 3.5) - Visual Basic .NET (Version 9.0) - 2009
' Autor ....:  - Aguila - Joaquin
' Emilio ...: mailto:joaquin@medina.name
'
' DESCRIPCION.:
'    Clase Base que implementa el patron Singleton
'
' OBSERVACIONES.:
'    Sin la referecia [Microsoft.VisualBasic] Sin Usar ControlChars
'    Utiliza reflexion para obtener el constructor de la clase que Hija que hereda esta clase
'    La clase hija NO debe tener un cosntructor public o friend
'
' BIBLIOGRAFIA.:
'    El Patrón Singleton (Por León Welicki )
'        http://msdn.microsoft.com/es-es/library/bb972272.aspx
'    Generic Singleton Pattern using Reflection, in C#
'        http://www.codeproject.com/KB/architecture/GenericSingletonPattern.aspx
'    Como implementar un Singleton concurrente - mayo 23, 2009 12:33 por Luis Guerrero
'        http://www.luisguerrero.net/post/2009/05/23/Como-implementar-un-Singleton-concurrente.aspx
'    Generic Singleton Pattern using Reflection, in C#
'        http://www.codeproject.com/KB/architecture/GenericSingletonPattern.aspx
'
'
' MODIFICACIONES [dd/mm/aaaa] 
' - 28/12/2008 10:39:12 - Creación 
' - 18/01/2009 - Incluyo el Atributo Serializable
' - 17/02/2009 - Creación en Util
' - 16/06/2009 - Modificada la Property [GetInstance]
' - 18/06/2009 - Incluyo la función     [IsInstaceCreada]
' - 16/01/2010 - Incluyo [TextoConFechaCreacionInstancia]
' - 02/01/2011 - Conversion de la clase incial en Generica
'              - Modificada la Property [GetInstance] con reflexion
' -------------------------------------------------
'
'*/
'


' -------------------------------------------------
Option Explicit On
Option Strict On
Option Infer Off
' Deshabilita la inferencia de tipo de variable local.
Option Compare Binary

''' <summary>
'''  Clase base singleton generica
''' </summary>
Public MustInherit Class SingletonBaseGeneric(Of T As {Class})
    Implements IDisposable


#Region "Parte Singleton"

    '-------------------------------------------------------------
    ' Para el objeto Singleton
    ' Regla FXCop (No inicializar innecesariamente)
    ' http://msdn2.microsoft.com/ms182274(VS.90).aspx 
    '-------------------------------------------------------------

    '------------------------------------
    ''' <summary>
    '''  Objeto interno que contendrá la instancia compartida de esta clase
    ''' </summary>
    Private Shared _SingletonObjetoInterno As T ' = Nothing

    '------------------------------------
    ''' <summary>
    '''  Fecha de creación de la instancia, solo para información (si es necesaria)
    ''' </summary>
    Private _SingletonFechaCreacion As DateTime ' = Nothing

    '------------------------------------
    ''' <summary>
    '''  El Mutex es una clase que proporciona .NET FrameWork
    '''  Con ella evitaremos que en un ambiente multi thread (multi hilos)
    '''  se cree algún tipo de conflicto
    ''' </summary>
    Private Shared _SingletonMutex As System.Threading.Mutex ' = Nothing

    '------------------------------------
    ''' <summary>
    ''' Constructor privado
    ''' </summary>
    ''' <remarks>
    ''' Protected para que se pueda heredar
    ''' Se inicializan las varibles internas con sus valores
    ''' </remarks>
    Protected Sub New()
        _SingletonFechaCreacion = DateTime.Now
        ' ojo no inicializar mutex (da error en el get instance)
    End Sub

    '------------------------------------
    ''' <summary>
    ''' Este método eliminará la instancia ya existente (si existe)
    ''' </summary>
    Public Shared Sub KillInstance()
        '---------------------------------------------------------
        ' No hago caso de la regla FxCop
        ' http://msdn.microsoft.com/library/ms182139(VS.100).aspx     
        ' Error, Certainty 95, for DoNotDeclareStaticMembersOnGenericTypes
        ' CA1000: No declarar miembros estáticos en tipos genéricos
        '---------------------------------------------------------
        If Not (_SingletonMutex Is Nothing) Then
            _SingletonMutex = Nothing
        End If

        If Not (_SingletonObjetoInterno Is Nothing) Then
            _SingletonObjetoInterno = Nothing
        End If
    End Sub

    '------------------------------------
    ''' <summary>
    ''' Este será el método que proporcionara el acceso 
    ''' a la instancia existente
    ''' </summary>
    ''' <returns>
    '''    <para>Devuelve la instancia de la clase</para>
    '''    <para>Si no existe, creara una nueva (método Singleton).</para>
    ''' </returns>
    ''' <exception cref="system.ObjectDisposedException">
    '''   <para> La instancia actual ya se ha eliminado.</para> 
    ''' </exception>
    ''' <exception cref="System.Threading.AbandonedMutexException">
    '''   <para> La espera finalizó porque un subproceso se cerró
    '''          sin liberar una exclusión mutua. </para>
    '''   <para> Esta excepción no se genera en Windows 98 ni en Windows Millennium.</para>
    ''' </exception>
    ''' <exception cref="System.ApplicationException">
    '''   <para> El subproceso que realiza la llamada no posee la exclusión mutua. </para> 
    ''' </exception>
    ''' <remarks> 
    ''' <para>Si no existe, creara una nueva (método Singleton).</para>
    ''' <para>Utilizo Mutex en lugar de Monitor, proque registra y 
    '''       controla automaticamente varias llamadas sin necesidad de
    '''       crear multiples objetos
    ''' </para>
    ''' </remarks>
    Public Shared Function GetInstance() As T
        '---------------------------------------------------------
        ' No hago caso de la regla FxCop
        ' http://msdn.microsoft.com/library/ms182139(VS.100).aspx     
        ' Error, Certainty 95, for DoNotDeclareStaticMembersOnGenericTypes
        ' CA1000: No declarar miembros estáticos en tipos genéricos
        '---------------------------------------------------------
        ' No hago caso de la regla FxCop
        'http://msdn.microsoft.com/library/ms182181(VS.100).aspx 
        'CA1024: Utilizar las propiedades donde corresponda
        '---------------------------------------------------------
        ' ATENCION
        ' sigo la regla 
        ' CA1065: No producir excepciones en ubicaciones inesperadas
        ' http://msdn.microsoft.com/library/bb386039(VS.100).aspx
        ' Las propiedades son campos inteligentes. Por consiguiente, 
        ' deberían comportarse de la forma más parecida a un campo.  
        ' Los campos no inician excepciones y las propiedades tampoco  
        ' deberían iniciarlas. Si usa una propiedad que inicia una  
        ' excepción, considere convertirla en un método. 
        '---------------------------------------------------------
        Try
            '----------------------------------------------------------------
            ' si no existe la instancia de esta clase crearla
            ' observa que utilizo dos veces esta pregunta
            ' la razon es que bloquear consume muchos recursos, de forma 
            ' que antes de boquear me aseguro de que es nothing.
            ' y una vez bloquedo antes de crear la instancia me vuelvo a
            ' asegurar de que no esta creada por algun otro proceso.
            If _SingletonObjetoInterno Is Nothing Then

                '-----------------------------------------------------
                ' Crear la instancia de la clase mutex (no existe)
                ' y ademas no debe existir
                Using _SingletonMutex As New System.Threading.Mutex

                    '-----------------------------------------------------
                    ' Utilizar reflexion para localizar el constructor de la clase (Of T)
                    Dim constructor As Reflection.ConstructorInfo = Nothing

                    ' Localizar el constructro privado de la clase hija
                    ' Utilizar Binding flags para excluir constructores publicos.
                    ' BindingFlags (Enumeración) Especifica los marcadores que controlan el enlace y 
                    ' la manera en que se realiza la búsqueda de miembros y tipos por reflexión. 
                    ' * Instance  Indica que se deben incluir los miembros de instancia en la búsqueda.  
                    ' * NonPublic  Indica que se deben incluir los miembros no públicos en la búsqueda.  
                    '-----------------------------------------------------
                    constructor = GetType(T).GetConstructor((Reflection.BindingFlags.Instance Or Reflection.BindingFlags.NonPublic), Nothing, Type.EmptyTypes, Nothing)

                    '-----------------------------------------------------
                    ' * si el constructor es Nothing o bien
                    ' * el constructor es visible como mucho para otros tipos del
                    '   mismo ensamblado y no es visible para los tipos derivados 
                    '   fuera del ensamblado. es decir es Public, Friend, o Protected Friend
                    '-----------------------------------------------------
                    If ((constructor Is Nothing) OrElse (constructor.IsAssembly)) Then
                        ' componer el mensaje de error
                        Dim mensajeError As String
                        mensajeError = String.Format( _
                            System.Globalization.CultureInfo.CurrentCulture, _
                            "No se encuentra un constructor 'Private' o 'Protected'  para el objeto '{0}'.", _
                            GetType(T).Name)
                        ' Escribir en el LOG
                        My.Application.Log.WriteEntry(mensajeError, TraceEventType.Error)
                        ' disparar exception
                        Throw New SingletonException(mensajeError)
                    End If

                    '-----------------------------------------------------
                    ' bloquea el subproceso actual 
                    _SingletonMutex.WaitOne()

                    '-----------------------------------------------------
                    ' si no existe (todavia) la instancia de esta clase crearla
                    '(usando el constructor (private o Protected)  obtenido por reflexion
                    If _SingletonObjetoInterno Is Nothing Then
                        _SingletonObjetoInterno = CType(constructor.Invoke(Nothing), T)
                    End If

                    '-----------------------------------------------------
                    ' desloquear el subproceso
                    _SingletonMutex.ReleaseMutex()

                    '-----------------------------------------------------
                    ' cargarme el objeto [Reflection.ConstructorInfo]
                    constructor = Nothing
                    ' cargarme la clase mutex que ya no la voy a volver a emplear
                    ' se hace con Using... End Using
                End Using

            End If ' (_SingletonObjetoInterno Is Nothing)

            '-----------------------------------------------------
            '-----------------------------------------------------
            ' devolver la instancia de este objeto
            ' despues del  End Try
            '-----------------------------------------------------
            '-----------------------------------------------------
            '-----------------------------------------------------
        Catch ex As System.ObjectDisposedException
            '   <para> La instancia actual ya se ha eliminado.</para> 
            ' componer el mensaje de error
            Dim mensajeError As String
            mensajeError = String.Format( _
                            System.Globalization.CultureInfo.CurrentCulture, _
                            " Objeto '{0}'. La instancia actual ya se ha eliminado.", GetType(T).Name)
            ' Escribir en el LOG
            My.Application.Log.WriteEntry(mensajeError, TraceEventType.Error)
            ' disparar exception
            Throw New SingletonException(mensajeError, ex)
            '-----------------------------------------------------

        Catch ex As System.Threading.AbandonedMutexException
            '   <para> La espera finalizó porque un subproceso se cerró  sin liberar una exclusión mutua. </para>
            '   <para> Esta excepción no se genera en Windows 98 ni en Windows Millennium.</para>
            ' componer el mensaje de error
            Dim mensajeError As String
            mensajeError = String.Format( _
                            System.Globalization.CultureInfo.CurrentCulture, _
                            " Objeto '{0}'. La espera finalizó porque un subproceso se cerró sin liberar una exclusión mutua.", GetType(T).Name)
            ' Escribir en el LOG
            My.Application.Log.WriteEntry(mensajeError, TraceEventType.Error)
            ' disparar exception
            Throw New SingletonException(mensajeError, ex)
            '-----------------------------------------------------

        Catch ex As System.ApplicationException
            '   <para> El subproceso que realiza la llamada no posee la exclusión mutua. </para> 
            ' componer el mensaje de error
            Dim mensajeError As String
            mensajeError = String.Format( _
                            System.Globalization.CultureInfo.CurrentCulture, _
                            " Objeto '{0}'. El subproceso que realiza la llamada no posee la exclusión mutua.", GetType(T).Name)
            ' Escribir en el LOG
            My.Application.Log.WriteEntry(mensajeError, TraceEventType.Error)
            ' disparar exception
            Throw New SingletonException(mensajeError, ex)
            '-----------------------------------------------------

        Catch ex As Exception
            ' componer el mensaje de error
            Dim mensajeError As String
            mensajeError = String.Format( _
                            System.Globalization.CultureInfo.CurrentCulture, _
                            " Objeto '{0}'. Problemas al generar el objeto singleton.", GetType(T).Name)
            mensajeError = mensajeError & Environment.NewLine
            ' Escribir en el LOG
            My.Application.Log.WriteEntry(mensajeError & ex.ToString, TraceEventType.Error)
            ' disparar exception
            Throw New SingletonException(mensajeError & ex.Message, ex)
            '-----------------------------------------------------
        End Try
        '-----------------------------------------------------
        ' devolver la instancia de este objeto
        Return _SingletonObjetoInterno
    End Function

    '------------------------------------
    ''' <summary>
    '''  Valor lógico que indica si la instancia de la clase esta creada o no
    ''' </summary>
    ''' <returns>
    '''     <para> Un valor lógico que indica:</para>
    '''     <para> TRUE -> La instancia de la clase esta creada</para>
    '''     <para> FALSE-> La instacia no esta creada</para>
    ''' </returns>
    Public Shared Function IsInstaceCreada() As Boolean
        '---------------------------------------------------------
        ' No hago caso de la regla FxCop
        ' http://msdn.microsoft.com/library/ms182139(VS.100).aspx     
        ' Error, Certainty 95, for DoNotDeclareStaticMembersOnGenericTypes
        ' CA1000: No declarar miembros estáticos en tipos genéricos
        '---------------------------------------------------------
        If _SingletonObjetoInterno Is Nothing Then
            Return False
        Else
            Return True
        End If
    End Function


    '------------------------------------
    ''' <summary>
    ''' Texto que informa de la fecha y hora en que fue creada la instancia
    ''' </summary>
    Public ReadOnly Property TextoConFechaCreacionInstancia() As String
        Get
            Return _SingletonFechaCreacion.ToShortDateString & _
                   " a las " & _
                   _SingletonFechaCreacion.ToLongTimeString & _
                   " horas"
        End Get
    End Property

    '------------------------------------
    ''' <summary>
    ''' Propiedad que presenta la fecha y hora en que fue creada la instancia
    ''' </summary>
    Public ReadOnly Property FechaCreacionInstancia() As Date
        Get
            Return _SingletonFechaCreacion
        End Get
    End Property
#End Region


#Region "IDisposable Support"
    Private disposedValue As Boolean ' Para detectar llamadas redundantes

    ' IDisposable
    Protected Overridable Sub Dispose(ByVal disposing As Boolean)
        If Not Me.disposedValue Then
            If disposing Then
                '----------------------------------------------------------
                ' Dispose de los recursos administrados
                '----------------------------------------------------------
                ' Liberar los recursos propios 
                '----------------------------------------------------------
                '  ---1--- Cerrar archivos abiertos  
                '          (No hay ninguno)

                '  -------------------------------------
                '  ---2--- Objetos Contenios a nothing
                '          [Objeto]=nothing

                'Objeto que contendrá la instancia compartida de esta clase
                If Not (_SingletonObjetoInterno Is Nothing) Then
                    _SingletonObjetoInterno = Nothing
                End If

                'El Mutex es una clase que proporciona .NET FrameWork
                If Not (_SingletonMutex Is Nothing) Then
                    _SingletonMutex = Nothing
                End If

                ' fecha de crecion, solo para informacion (si es necesaria)
                If Not (_SingletonFechaCreacion = Nothing) Then
                    _SingletonFechaCreacion = Nothing
                End If
            End If
        End If
        Me.disposedValue = True
    End Sub

    ' invalidar Finalize() sólo si la instrucción Dispose(ByVal disposing As Boolean) 
    '  anterior tiene código para liberar recursos no administrados.
    ' Apunte Tactico:
    '  Un objeto no administrado es con lo que hemos trabajado toda la vida antes de .NET, 
    '  es cualquier objeto ajeno a .NET, por ejemplo, un objeto creado directamente 
    '  llamando al API de Windows (por ejemplo el HWND de una ventana del sistema) 
    'Protected Overrides Sub Finalize()
    '    ' No cambie este código. Ponga el código de limpieza en la instrucción Dispose(ByVal disposing As Boolean) anterior.
    '    Dispose(False)
    '    MyBase.Finalize()
    'End Sub

    ' Visual Basic agregó este código para implementar correctamente el modelo descartable.
    Public Sub Dispose() Implements IDisposable.Dispose
        ' No cambie este código. Coloque el código de limpieza en Dispose (ByVal que se dispone como Boolean).
        Dispose(True)
        GC.SuppressFinalize(Me)
    End Sub
#End Region

End Class





''' <summary>
'''  Representa un error que ocurre al crear un objeto Singletonsingleton.
''' </summary>
''' <remarks>
''' http://msdn.microsoft.com/en-us/library/ms229064(VS.80).aspx
''' </remarks>
<Serializable()> _
Public Class SingletonException
    Inherits Exception

    ''' <summary>
    ''' Constructor
    ''' </summary>
    Public Sub New()
        MyBase.New()
    End Sub

    ''' <summary>
    ''' Inicializa una nueva instancia con un mensaje de error
    ''' </summary>
    ''' <param name="message">El mensaje que describe el error.</param>
    Public Sub New(ByVal message As String)
        MyBase.New(message)

    End Sub

    ''' <summary>
    ''' Inicializa una nueva instancia con una referencia
    ''' a la excepcion que es causa de esta excepcion
    ''' </summary>
    ''' <param name="innerException">
    ''' La escepcion que es el origen de que se cree esta excepcion
    ''' o una referencia (null)
    ''' </param>
    Public Sub New(ByVal innerException As Exception)
        MyBase.New(Nothing, innerException)
    End Sub

    ''' <summary>
    ''' Inicializa una instancia con un mensaje de error especifico
    ''' y una referencia al objeto Excepcion que es la causa de esta excepcion
    ''' </summary>
    ''' <param name="message">El mensaje que describe el error.</param>
    ''' <param name="innerException">
    ''' La escepcion que es el origen de que se cree esta excepcion
    ''' o una referencia (null)
    ''' </param>
    Public Sub New(ByVal message As String, ByVal innerException As Exception)
        MyBase.New(message, innerException)
    End Sub



    ''' <summary>
    '''   Constructor de serializacion
    '''   Inicializa una nueva instancia de la clase a partir de datos serializados. 
    ''' </summary>
    ''' <param name="info">Objeto que contiene los datos serializados.</param>
    ''' <param name="context">nformación contextual acerca del origen o el destino de la excepción.</param>
    ''' <remarks><code>http://msdn.microsoft.com/es-ES/library/ms182343.aspx</code></remarks>
    Protected Sub New(ByVal info As Runtime.Serialization.SerializationInfo, ByVal context As Runtime.Serialization.StreamingContext)
        '---------------------------------------------------------
        ' ¡ATENCION!  regla FxCop
        'CA2229: Implementar constructores de serialización
        'http://msdn.microsoft.com/es-ES/library/ms182343.aspx
        'CA1032: Implementar constructores de excepción estándar
        'http://msdn.microsoft.com/library/ms182151(VS.100).aspx
        '---------------------------------------------------------
        MyBase.New(info, context)
    End Sub


End Class


